/****************************************************************************
Copyright (c) 2012,Otto-von-Guericke University of Magdeburg. All rights 
reserved.

This program and accompanying materials are made available under the terms of
the Eclipse Public License - v 1.0, http://www.eclipse.org/legal/epl-v10.html
 ****************************************************************************/

package index.structures;

import index.base.AIndexStructure;
import metrics.DistanceMetric;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;

import javax.swing.JCheckBox;
import javax.swing.JFormattedTextField;
import javax.swing.JFrame;
import javax.swing.JPanel;

import testapp.Dictionary;
import testapp.Dictionary.Bundle;
import testapp.JOptionDialog;
import testapp.tapedpane.JTappedPane_Panel;

public class IndexP_StableLSH extends AIndexStructure {
	private int numberOfHashFunktions = 10;
	private int W = 10;
	private int seed = 10;
	private boolean useP2Stable = false;

	public IndexP_StableLSH(int size, int dim, int minValue, int maxValue,
			HashMap<String, Object> map) {
		super(size, dim, minValue, maxValue, map);
		initParameter(map);

		rand = new Random(seed);
		generateHashFunctions();
	}

	private double getP1Stable(double x) {
		return (1 / Math.PI) * (1 / (1 + x * x));
	}

	private double getP2Stable(double x) {
		return (1 / Math.sqrt(2 * Math.PI) * Math.exp(-(x * x) / 2));

	}

	double[][] aVectorsForHashFunctions;
	int[] bForHashFunctions;

	HashMap<Integer, ArrayList<Integer>>[] hashMaps;

	Random rand;

	// int[][] allPoints;

	@SuppressWarnings("unchecked")
	private void generateHashFunctions() {
		hashMaps = new HashMap[numberOfHashFunktions];

		aVectorsForHashFunctions = new double[numberOfHashFunktions][dim];
		bForHashFunctions = new int[numberOfHashFunktions];
		for (int hash = 0; hash < aVectorsForHashFunctions.length; hash++) {
			for (int dim = 0; dim < aVectorsForHashFunctions[0].length; dim++) {
				double value = rand.nextDouble() * W;
				if (useP2Stable) {
					aVectorsForHashFunctions[hash][dim] = getP2Stable(value);
				} else {
					aVectorsForHashFunctions[hash][dim] = getP1Stable(value);
				}
			}
			bForHashFunctions[hash] = (int) rand.nextDouble() * W;

			hashMaps[hash] = new HashMap<Integer, ArrayList<Integer>>();
		}
	}

	private int getHashCode(int[] point, int hashFunc) {
		long t = 0;
		double[] currHFunc = aVectorsForHashFunctions[hashFunc];

		for (int i = 0; i < point.length; i++) {
			t += point[i] * currHFunc[i];
		}
		t += bForHashFunctions[hashFunc];
		return (int) t / W;
	}

	private int[] getHashVector(int[] point) {
		int[] hashedVec = new int[numberOfHashFunktions];
		for (int i = 0; i < hashedVec.length; i++) {
			hashedVec[i] = getHashCode(point, i);
		}
		return hashedVec;
	}

	public boolean canKNNSearch() {
		return true;
	}

	public void clear() {
		aVectorsForHashFunctions = null;
		bForHashFunctions = null;
		hashMaps = null;
	}

	@Override
	public String getInformations() {
		String pStable = "p1";
		if (useP2Stable) {
			pStable = "p2";
		}
		return getName() + " \n " + Dictionary.get(Bundle.INDEX, "numberhash")
				+ numberOfHashFunktions + "\nW: " + W + "\n"
				+ Dictionary.get(Bundle.INDEX, "stable") + pStable + "-stable";
	}

	@Override
	public JOptionDialog getOptionsDialog(JFrame parent) {
		return new PStabLSHOptionsDialog(parent, this);
	}

	int alreadyInsertet = 0;

	public boolean insert(int[] dataset) {
		int[] hashValues = getHashVector(dataset);

		for (int i = 0; i < hashValues.length; i++) {
			HashMap<Integer, ArrayList<Integer>> currMap = hashMaps[i];
			if (!currMap.containsKey(hashValues[i])) {
				ArrayList<Integer> hashMap = new ArrayList<Integer>();
				hashMap.add(alreadyInsertet);
				currMap.put(hashValues[i], hashMap);
			} else {
				ArrayList<Integer> list = currMap.get(hashValues[i]);
				list.add(alreadyInsertet);
			}
		}

		alreadyInsertet++;

		return true;
	}

	@Override
	public int[] search(int[] datasetToSearch) {
		int[] hashCode = getHashVector(datasetToSearch);
		HashMap<Integer, ArrayList<Integer>> map = hashMaps[0];
		ArrayList<Integer> list = map.get(hashCode[0]);
		if (list == null)
			return null;
		for (int i = 0; i < list.size(); i++) {
			Integer currIndex = list.get(i);
			int[] currP = getDataset(currIndex);

			boolean isIdentity = true;
			for (int k = 0; k < currP.length; k++) {
				if (currP[k] != datasetToSearch[k]) {
					isIdentity = false;
					break;
				}
			}
			if (isIdentity)
				return currP;

			// if(metrik.calculateDistance(datasetToSearch, currP) == 0){
			// return currP;
			// }
		}
		return null;
	}

	public int[][] searchKNN(int[] datasetToSearch, int k) {
		HashMap<Integer, int[]> possibleKNN = new HashMap<Integer, int[]>();

		boolean[] pointFound = new boolean[this.size];
		int[][] result = new int[k][this.dim];// return value
		int[] candIndex = new int[k];// Candidates: the index of the point in
										// datasets[size]
		double[] canDistances = new double[k];// belongs to *index[]*, contains
												// corresponding distance of
												// this point to
												// *datasetToSearch*
		double maxDistance = Double.MAX_VALUE;
		int indexMaxDistance = 0;

		// (1) initialize array with 0, (Long.MAX_VALUE)
		for (int i = 0; i < k; i++) {
			candIndex[i] = 0;
			canDistances[i] = Double.MAX_VALUE;
		}

		// (2)For each hash function
		int[] hashCode = getHashVector(datasetToSearch);
		for (int curHashFunc = 0; curHashFunc < hashCode.length; curHashFunc++) {
			HashMap<Integer, ArrayList<Integer>> map = hashMaps[curHashFunc];

			// (2a) get Bucket if existing
			if (map.containsKey(hashCode[curHashFunc])) {
				ArrayList<Integer> currList = map.get(hashCode[curHashFunc]);

				// (2b) loop over points in the bucket
				for (int pointInBucket = 0; pointInBucket < currList.size(); pointInBucket++) {
					int currIndex = currList.get(pointInBucket);

					// (2c) consider adding this point only if not already added
					if (!pointFound[currIndex]) {
						pointFound[currIndex] = true;

						// Now the part, which the same as in the sequential
						// scan
						// add only if *dist* smaller than *maxDistance*
						int[] data = getDataset(currIndex);
						possibleKNN.put(currIndex, data);
						double dist = getMetric().calculateDistance(
								datasetToSearch, data);
						if (dist < maxDistance) {
							// exchange *one* point containing the max dist
							candIndex[indexMaxDistance] = currIndex;
							canDistances[indexMaxDistance] = dist;

							// determine new *maxDistance*
							maxDistance = Double.MIN_VALUE;
							for (int i = 0; i < k; i++) {
								if (maxDistance < canDistances[i]) {
									maxDistance = canDistances[i];
									indexMaxDistance = i;
								}
							}
						}
					}
				}
			}
		}

		// (3) Finally, sort array and get the real points, because its
		// unordered and the order shows the distances ascending
		int stillToSort = k - 1;// Because arrays start with [0]
		while (stillToSort > 0) {// do it k-times

			// Here we copy an element having maxDist to the end of (the
			// unsorted part) of the arrays -> i.e., distances[stillToSort]
			double tempDist = canDistances[stillToSort];
			int tempIndex = candIndex[stillToSort];
			// now copy a point having *maxDist* to the end of the array
			canDistances[stillToSort] = canDistances[indexMaxDistance];
			candIndex[stillToSort] = candIndex[indexMaxDistance];
			// copy to the end of unordered array
			canDistances[indexMaxDistance] = tempDist;
			candIndex[indexMaxDistance] = tempIndex;

			// get the real point-> Don't want to use an extra for loop in the
			// end
			result[stillToSort] = possibleKNN.get(candIndex[stillToSort]);

			// determine new *maxDist* ignoring the recently swaped element
			maxDistance = Double.MIN_VALUE;
			for (int i = 0; i < stillToSort; i++) {// the element
													// (distances[stillToSort])
													// we recently swaped is not
													// considered!
				if (maxDistance < canDistances[i]) {
					maxDistance = canDistances[i];
					indexMaxDistance = i;
				}
			}
			stillToSort--;// consider only the unordered part of the array
		}
		// because of while(stillToSort>0)
		result[0] = possibleKNN.get(candIndex[0]);

		return result;
	}

	private class PStabLSHOptionsDialog extends JOptionDialog {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		JFormattedTextField fieldNumberFunctions;
		JFormattedTextField fieldW;
		JFormattedTextField fieldSeed;
		JCheckBox fieldUseP2Stable;

		JTappedPane_Panel northPanel;

		public PStabLSHOptionsDialog(JFrame parent, AIndexStructure is) {
			super(parent, is.getName());
		}

		public JPanel getNorthPanel() {
			northPanel = new JTappedPane_Panel(null) {
				private static final long serialVersionUID = 1L;

				@Override
				public void refreshResource() {

				}

			};

			fieldNumberFunctions = northPanel
					.getFormtedTextFieldInstance_Integer(10);
			fieldNumberFunctions.setValue(10);
			fieldNumberFunctions.setColumns(3);

			northPanel.addSimpleLabelAndInputRow(0,
					Dictionary.get(Bundle.INDEX, "numberhash") + ":",
					fieldNumberFunctions);

			fieldW = northPanel.getFormtedTextFieldInstance_Integer(10);
			fieldW.setValue(10);
			fieldW.setColumns(3);
			fieldW.setToolTipText(Dictionary.get(Bundle.INDEX, "wtooltip"));
			northPanel.addSimpleLabelAndInputRow(1, "W", fieldW);

			fieldSeed = northPanel.getFormtedTextFieldInstance_Integer(10);
			fieldSeed.setValue(10);
			fieldSeed.setColumns(3);
			northPanel.addSimpleLabelAndInputRow(2,
					Dictionary.get(Bundle.INDEX, "randomseed"), fieldSeed);

			fieldUseP2Stable = new JCheckBox("", false);
			fieldUseP2Stable.setToolTipText(Dictionary.get(Bundle.INDEX,
					"pstabletooltip"));
			northPanel.addSimpleLabelAndInputRow(3,
					Dictionary.get(Bundle.INDEX, "pstableinfo") + ":",
					fieldUseP2Stable);

			return northPanel;
		}

		@Override
		public HashMap<String, Object> getOptions() {
			HashMap<String, Object> map = new HashMap<String, Object>();
			map.put("numberOfHashFunktions", ""
					+ ((Number) fieldNumberFunctions.getValue()).intValue());
			map.put("W", "" + ((Number) fieldW.getValue()).intValue());
			map.put("seed", "" + ((Number) fieldSeed.getValue()).intValue());
			map.put("useP2Stable", "" + fieldUseP2Stable.isSelected());
			return map;
		}
	}

	@Override
	public String getName() {
		return Dictionary.get(Bundle.INDEX, "lsh");
	}

}
